@@ -29,6 +29,8 @@ from stevedore import driver
 
 import tooz
 from tooz import _retry
+from tooz import partitioner
+from tooz import utils
 
 LOG = logging.getLogger(__name__)
 
@@ -127,6 +129,11 @@ class MemberJoinedGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: +member %s>" % (self.__class__.__name__,
+                                               self.member_id,
+                                               self.group_id)
+
 
 class MemberLeftGroup(Event):
     """A member left a group event."""
@@ -135,6 +142,11 @@ class MemberLeftGroup(Event):
         self.group_id = group_id
         self.member_id = member_id
 
+    def __repr__(self):
+        return "<%s: group %s: -member %s>" % (self.__class__.__name__,
+                                               self.member_id,
+                                               self.group_id)
+
 
 class LeaderElected(Event):
     """A leader as been elected."""
@@ -215,7 +227,6 @@ class Heart(object):
         return self._finished.is_set()
 
 
-@six.add_metaclass(abc.ABCMeta)
 class CoordinationDriver(object):
 
     requires_beating = False
@@ -232,21 +243,55 @@ class CoordinationDriver(object):
     enum member(s) that can be used to interogate how this driver works.
     """
 
-    def __init__(self):
+    def __init__(self, member_id):
+        super(CoordinationDriver, self).__init__()
+        self._member_id = member_id
         self._started = False
         self._hooks_join_group = collections.defaultdict(Hooks)
         self._hooks_leave_group = collections.defaultdict(Hooks)
         self._hooks_elected_leader = collections.defaultdict(Hooks)
-        # A cache for group members
-        self._group_members = collections.defaultdict(set)
         self.requires_beating = (
             CoordinationDriver.heartbeat != self.__class__.heartbeat
         )
         self.heart = Heart(self)
 
     def _has_hooks_for_group(self, group_id):
-        return (len(self._hooks_join_group[group_id]) +
-                len(self._hooks_leave_group[group_id]))
+        return (group_id in self._hooks_join_group or
+                group_id in self._hooks_leave_group)
+
+    def join_partitioned_group(
+            self, group_id,
+            weight=1,
+            partitions=partitioner.Partitioner.DEFAULT_PARTITION_NUMBER):
+        """Join a group and get a partitioner.
+
+        A partitioner allows to distribute a bunch of objects across several
+        members using a consistent hash ring. Each object gets assigned (at
+        least) one member responsible for it. It's then possible to check which
+        object is owned by any member of the group.
+
+        This method also creates if necessary, and joins the group with the
+        selected weight.
+
+        :param group_id: The group to create a partitioner for.
+        :param weight: The weight to use in the hashring for this node.
+        :param partitions: The number of partitions to create.
+        :return: A :py:class:`~tooz.partitioner.Partitioner` object.
+
+        """
+        self.join_group_create(
+            group_id, capabilities=utils.dumps({'weight': weight}))
+        return partitioner.Partitioner(self, group_id)
+
+    def leave_partitioned_group(self, partitioner):
+        """Leave a partitioned group.
+
+        This leaves the partitioned group and stop the partitioner.
+        :param group_id: The group to create a partitioner for.
+        """
+        leave = self.leave_group(partitioner.group_id)
+        partitioner.stop()
+        return leave.get()
 
     @staticmethod
     def run_watchers(timeout=None):
@@ -262,7 +307,6 @@ class CoordinationDriver(object):
         """Try to leader elect this coordinator & activate hooks on success."""
         raise tooz.NotImplemented
 
-    @abc.abstractmethod
     def watch_join_group(self, group_id, callback):
         """Call a function when group_id sees a new member joined.
 
@@ -275,7 +319,6 @@ class CoordinationDriver(object):
         """
         self._hooks_join_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_join_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member joined.
 
@@ -284,15 +327,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_join_group:
+                raise ValueError
             self._hooks_join_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
+        if not self._hooks_join_group[group_id]:
+            del self._hooks_join_group[group_id]
 
-    @abc.abstractmethod
     def watch_leave_group(self, group_id, callback):
         """Call a function when group_id sees a new member leaving.
 
@@ -306,7 +351,6 @@ class CoordinationDriver(object):
         """
         self._hooks_leave_group[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_leave_group(self, group_id, callback):
         """Stop executing a function when a group_id sees a new member leaving.
 
@@ -315,18 +359,17 @@ class CoordinationDriver(object):
                          this group
         """
         try:
+            # Check if group_id is in hooks to avoid creating a default empty
+            # entry in hooks list.
+            if group_id not in self._hooks_leave_group:
+                raise ValueError
             self._hooks_leave_group[group_id].remove(callback)
         except ValueError:
             raise WatchCallbackNotFound(group_id, callback)
 
-        if (not self._has_hooks_for_group(group_id) and
-           group_id in self._group_members):
-            del self._group_members[group_id]
-
         if not self._hooks_leave_group[group_id]:
             del self._hooks_leave_group[group_id]
 
-    @abc.abstractmethod
     def watch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -340,7 +383,6 @@ class CoordinationDriver(object):
         """
         self._hooks_elected_leader[group_id].append(callback)
 
-    @abc.abstractmethod
     def unwatch_elected_as_leader(self, group_id, callback):
         """Call a function when member gets elected as leader.
 
@@ -379,7 +421,7 @@ class CoordinationDriver(object):
         is initiated.
         """
         if self._started:
-            raise ToozError(
+            raise tooz.ToozError(
                 "Can not start a driver which has not been stopped")
         self._start()
         if self.requires_beating and start_heart:
@@ -398,16 +440,19 @@ class CoordinationDriver(object):
         disappear from all joined groups.
         """
         if not self._started:
-            raise ToozError("Can not stop a driver which has not been started")
+            raise tooz.ToozError(
+                "Can not stop a driver which has not been started")
         if self.heart.is_alive():
             self.heart.stop()
             self.heart.wait()
-        leaving = [self.leave_group(group)
-                   for group in self._joined_groups]
-        for leave in leaving:
+        # Some of the drivers modify joined_groups when being called to leave
+        # so clone it so that we aren't modifying something while iterating.
+        joined_groups = self._joined_groups.copy()
+        leaving = [self.leave_group(group) for group in joined_groups]
+        for fut in leaving:
             try:
-                leave.get()
-            except ToozError:
+                fut.get()
+            except tooz.ToozError:
                 # Whatever happens, ignore. Maybe we got booted out/never
                 # existed in the first place, or something is down, but we just
                 # want to call _stop after whatever happens to not leak any
@@ -453,7 +498,7 @@ class CoordinationDriver(object):
         raise tooz.NotImplemented
 
     @_retry.retry()
-    def join_group_create(self, group_id):
+    def join_group_create(self, group_id, capabilities=b""):
         """Join a group and create it if necessary.
 
         If the group cannot be joined because it does not exist, it is created
@@ -464,9 +509,9 @@ class CoordinationDriver(object):
         if another member is creating/deleting the group at the same time.
 
         :param group_id: Identifier of the group to join and create
-
+        :param capabilities: the capabilities of the joined member
         """
-        req = self.join_group(group_id)
+        req = self.join_group(group_id, capabilities)
         try:
             req.get()
         except GroupNotCreated:
@@ -606,18 +651,60 @@ class CoordAsyncResult(object):
         """Returns True if the task is done, False otherwise."""
 
 
-class _RunWatchersMixin(object):
-    """Mixin to share the *mostly* common ``run_watchers`` implementation."""
+class CoordinationDriverCachedRunWatchers(CoordinationDriver):
+    """Coordination driver with a `run_watchers` implementation.
+
+    This implementation of `run_watchers` is based on a cache of the group
+    members between each run of `run_watchers` that is being updated between
+    each run.
+
+    """
+
+    def __init__(self, member_id):
+        super(CoordinationDriverCachedRunWatchers, self).__init__(member_id)
+        # A cache for group members
+        self._group_members = collections.defaultdict(set)
+        self._joined_groups = set()
+
+    def _init_watch_group(self, group_id):
+        members = self.get_members(group_id).get()
+        if group_id not in self._group_members:
+            self._group_members[group_id] = members
+
+    def watch_join_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_join_group(
+            group_id, callback)
+
+    def unwatch_join_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_join_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
+
+    def watch_leave_group(self, group_id, callback):
+        self._init_watch_group(group_id)
+        super(CoordinationDriverCachedRunWatchers, self).watch_leave_group(
+            group_id, callback)
+
+    def unwatch_leave_group(self, group_id, callback):
+        super(CoordinationDriverCachedRunWatchers, self).unwatch_leave_group(
+            group_id, callback)
+
+        if (not self._has_hooks_for_group(group_id) and
+           group_id in self._group_members):
+            del self._group_members[group_id]
 
     def run_watchers(self, timeout=None):
         with timeutils.StopWatch(duration=timeout) as w:
-            known_groups = self.get_groups().get(
-                timeout=w.leftover(return_none=True))
             result = []
-            for group_id in known_groups:
+            group_with_hooks = set(self._hooks_join_group.keys()).union(
+                set(self._hooks_leave_group.keys()))
+            for group_id in group_with_hooks:
                 try:
-                    group_members_fut = self.get_members(group_id)
-                    group_members = group_members_fut.get(
+                    group_members = self.get_members(group_id).get(
                         timeout=w.leftover(return_none=True))
                 except GroupNotCreated:
                     group_members = set()
@@ -684,40 +771,29 @@ def get_coordinator(backend_url, member_id,
     return d
 
 
-class ToozError(Exception):
-    """Exception raised when an internal error occurs.
-
-    Raised for instance in case of server internal error.
-
-    :ivar cause: the cause of the exception being raised, when not none this
-                 will itself be an exception instance, this is useful for
-                 creating a chain of exceptions for versions of python where
-                 this is not yet implemented/supported natively.
-
-    """
-
-    def __init__(self, message, cause=None):
-        super(ToozError, self).__init__(message)
-        self.cause = cause
+# TODO(harlowja): We'll have to figure out a way to remove this 'alias' at
+# some point in the future (when we have a better way to tell people it has
+# moved without messing up their exception catching hierarchy).
+ToozError = tooz.ToozError
 
 
-class ToozDriverChosenPoorly(ToozError):
+class ToozDriverChosenPoorly(tooz.ToozError):
     """Raised when a driver does not match desired characteristics."""
 
 
-class ToozConnectionError(ToozError):
+class ToozConnectionError(tooz.ToozError):
     """Exception raised when the client cannot connect to the server."""
 
 
-class OperationTimedOut(ToozError):
+class OperationTimedOut(tooz.ToozError):
     """Exception raised when an operation times out."""
 
 
-class LockAcquireFailed(ToozError):
+class LockAcquireFailed(tooz.ToozError):
     """Exception raised when a lock acquire fails in a context manager."""
 
 
-class GroupNotCreated(ToozError):
+class GroupNotCreated(tooz.ToozError):
     """Exception raised when the caller request an nonexistent group."""
     def __init__(self, group_id):
         self.group_id = group_id
@@ -725,7 +801,7 @@ class GroupNotCreated(ToozError):
             "Group %s does not exist" % group_id)
 
 
-class GroupAlreadyExist(ToozError):
+class GroupAlreadyExist(tooz.ToozError):
     """Exception raised trying to create an already existing group."""
     def __init__(self, group_id):
         self.group_id = group_id
@@ -733,7 +809,7 @@ class GroupAlreadyExist(ToozError):
             "Group %s already exists" % group_id)
 
 
-class MemberAlreadyExist(ToozError):
+class MemberAlreadyExist(tooz.ToozError):
     """Exception raised trying to join a group already joined."""
     def __init__(self, group_id, member_id):
         self.group_id = group_id
@@ -743,7 +819,7 @@ class MemberAlreadyExist(ToozError):
             (member_id, group_id))
 
 
-class MemberNotJoined(ToozError):
+class MemberNotJoined(tooz.ToozError):
     """Exception raised trying to access a member not in a group."""
     def __init__(self, group_id, member_id):
         self.group_id = group_id
@@ -752,14 +828,14 @@ class MemberNotJoined(ToozError):
                                               (member_id, group_id))
 
 
-class GroupNotEmpty(ToozError):
+class GroupNotEmpty(tooz.ToozError):
     "Exception raised when the caller try to delete a group with members."
     def __init__(self, group_id):
         self.group_id = group_id
         super(GroupNotEmpty, self).__init__("Group %s is not empty" % group_id)
 
 
-class WatchCallbackNotFound(ToozError):
+class WatchCallbackNotFound(tooz.ToozError):
     """Exception raised when unwatching a group.
 
     Raised when the caller tries to unwatch a group with a callback that
@@ -774,29 +850,7 @@ class WatchCallbackNotFound(ToozError):
             (callback.__name__, group_id))
 
 
-class SerializationError(ToozError):
-    "Exception raised when serialization or deserialization breaks."
-
-
-def raise_with_cause(exc_cls, message, *args, **kwargs):
-    """Helper to raise + chain exceptions (when able) and associate a *cause*.
-
-    **For internal usage only.**
-
-    NOTE(harlowja): Since in py3.x exceptions can be chained (due to
-    :pep:`3134`) we should try to raise the desired exception with the given
-    *cause*.
-
-    :param exc_cls: the :py:class:`~tooz.coordination.ToozError` class
-                    to raise.
-    :param message: the text/str message that will be passed to
-                    the exceptions constructor as its first positional
-                    argument.
-    :param args: any additional positional arguments to pass to the
-                 exceptions constructor.
-    :param kwargs: any additional keyword arguments to pass to the
-                   exceptions constructor.
-    """
-    if not issubclass(exc_cls, ToozError):
-        raise ValueError("Subclass of tooz error is required")
-    excutils.raise_with_cause(exc_cls, message, *args, **kwargs)
+# TODO(harlowja,jd): We'll have to figure out a way to remove this 'alias' at
+# some point in the future (when we have a better way to tell people it has
+# moved without messing up their exception catching hierarchy).
+SerializationError = utils.SerializationError
